"""Motion calculation from IMU data."""

import time
from copy import copy

import numpy as np

from py_imu.quaternion import Quaternion, Vector3D
from py_imu.utilities import RunningAverage, gravity, sensorAcc

MIN_MOTION_TIME = 0.5  # seconds, need to have had at least half of second motion to update velocity bias
HEADING_AVG_HISTORY = 5


class Motion:
    """IMU Motion Estimation."""

    def __init__(self, **kwargs):
        # Default values are for Tucson, Arizona, USA
        # This will is needed to compute magnitude of gravity
        #   which is subtracted from measured acceleration
        self.latitude = kwargs.get("latitude", 32.253460)  # decimal degrees
        self.altitude = kwargs.get("altitude", 730)  # meter

        self.m_heading_X = RunningAverage(HEADING_AVG_HISTORY)
        self.m_heading_Y = RunningAverage(HEADING_AVG_HISTORY)

        self.residuals_bias = Vector3D(0, 0, 0)  # on  the device frame coordinates

        self.worldResiduals = Vector3D(0, 0, 0)
        self.worldResiduals_previous = Vector3D(0, 0, 0)

        self.worldVelocity = Vector3D(0, 0, 0)
        self.worldVelocity_drift = Vector3D(0, 0, 0)  # in world corrdinate system
        self.world_velocity_previous = Vector3D(0, 0, 0)

        self.worldPosition = Vector3D(0, 0, 0)
        self.worldPosition_previous = Vector3D(0, 0, 0)

        self.driftLearningAlpha = 0.01  # Poorman's low-pass filter

        self.heading_X_avg = 0.0
        self.heading_Y_avg = 0.0
        self.m_heading = 0.0  # average heading in radians

        self.motion = False  # is the device moving?
        self.motion_previous = False
        self.motionStart_time = time.perf_counter()

        self.timestamp_previous = -1.0
        self.dt_motion = (
            0.0  # no motion has occurred yet, length of motion period in seconds
        )
        self.dt = 0.0

        self.localGravity = gravity(
            latitude=self.latitude, altitude=self.altitude
        )  # Gravity on Earth's (ellipsoid) Surface

    def reset(self):
        """Reset Motion Estimation."""
        self.residuals_bias = Vector3D(0, 0, 0)
        self.worldVelocity = Vector3D(0, 0, 0)
        self.worldPosition = Vector3D(0, 0, 0)
        self.world_velocity_previous = Vector3D(0, 0, 0)
        self.worldVelocity_drift = Vector3D(0, 0, 0)
        self.motion = False

    def resetPosition(self):
        """Reset Position."""
        self.worldPosition = Vector3D(0, 0, 0)
        self.worldPosition_previous = Vector3D(0, 0, 0)

    def updateAverageHeading(self, heading) -> float:
        """Update Average Heading."""
        ## this needs two component because of 0 - 360 jump at North
        self.m_heading_X.update(np.cos(heading))
        self.m_heading_Y.update(np.sin(heading))
        self.m_heading = np.atan2(self.m_heading_Y.avg, self.m_heading_X.avg)
        if self.m_heading < 0:
            self.m_heading += 2 * np.pi
        return self.m_heading

    def update(self, q: Quaternion, acc: Vector3D, moving: bool, timestamp: float):
        """Update Motion Estimation."""
        # Input:
        #  Quaternion
        #  Timestamp
        # Calculates:
        #  Acceleration in world coordinate system
        #  Velocity in world coordinate system
        #  Position in world coordinate system
        # If there is no motion
        #  Acceleration bias is updated
        #  Velocity bias is updated

        # Integration Time Step
        if self.timestamp_previous < 0.0:
            dt = 0.0
        else:
            dt = timestamp - self.timestamp_previous
            if dt > 1.0:
                # we had reconnection, avoid updating with large dt
                dt = 0.0
        self.timestamp_previous = copy(timestamp)

        # Acceleration residuals on the sensor
        self.residuals = sensorAcc(acc=acc, q=q, g=self.localGravity)
        self.residuals -= self.residuals_bias

        # Acceleration residuals in world coordinate system
        # self.worldResiduals = (q * self.residuals * q.conjugate).v
        self.worldResiduals = self.residuals.rotate(q.r33)

        # Motion Status, ongoing, no motion, ended?
        motion_ended = False
        if not self.motion_previous:
            if moving:
                # Motion Started
                self.motionStart_time = copy(timestamp)
        elif not moving:
            # Motion Ended
            motion_ended = True
        # Keep track of previous status
        self.motion_previous = copy(moving)

        # Update Velocity and Position when moving
        if moving:
            # Integrate acceleration and add to velocity (uses trapezoidal integration technique
            self.worldVelocity = self.world_velocity_previous + (
                (self.worldResiduals + self.worldResiduals_previous) * 0.5 * dt
            )
            self.dt_motion += dt

            # Update Velocity
            self.worldVelocity = self.worldVelocity - (self.worldVelocity_drift * dt)

            # Integrate velocity and add to position
            self.worldPosition = (
                self.worldPosition_previous
                + (self.worldVelocity + self.world_velocity_previous) * 0.5 * dt
            )

            # keep a history of previous values
            self.worldResiduals_previous = copy(self.worldResiduals)
            self.world_velocity_previous = copy(self.worldVelocity)
            self.worldPosition_previous = copy(self.worldPosition)

        else:  # no Motion
            # Estimate Velocity Bias when not moving
            # When motion ends, velocity should be zero
            if motion_ended and (self.dt_motion > MIN_MOTION_TIME):
                self.worldVelocity_drift = (
                    self.worldVelocity_drift * (1.0 - self.driftLearningAlpha)
                ) + ((self.worldVelocity / self.dt_motion) * self.driftLearningAlpha)
                self.dt_motion = 0.0

            # Reset Residuals
            self.worldResiduals_previous = Vector3D(x=0.0, y=0.0, z=0.0)
            # Reset Velocity
            self.worldVelocity = Vector3D(x=0.0, y=0.0, z=0.0)
            self.world_velocity_previous = Vector3D(x=0.0, y=0.0, z=0.0)

            # Update acceleration bias, when not moving residuals should be zero
            # If the accelerometer is not calibrated properly, subtracting bias will cause drift
            self.residuals_bias = (
                self.residuals_bias * (1.0 - self.driftLearningAlpha)
            ) + (self.residuals * self.driftLearningAlpha)

        self.dt = dt



"""Utility functions for the py_imu filter."""

import math
import numbers
import struct
from collections import deque
from copy import copy

import numpy as np

from py_imu.quaternion import Quaternion, Vector3D

IDENTITY_QUATERNION = Quaternion(1.0, 0.0, 0.0, 0.0)
VECTOR_ZERO = Vector3D(0.0, 0.0, 0.0)

###########################################################
# Utility Functions
###########################################################


def clip(val, largest):
    """Clip val to [0,largest]."""
    return 0 if val < 0 else largest if val > largest else val


def clamp(val, smallest, largest):
    """Clip val to [smallest, largest]."""
    if val < smallest:
        return smallest
    if val > largest:
        return largest
    return val


def asin(value: float):
    """Asin with clamping."""
    if value <= -1.0:
        return -math.pi / 2
    elif value >= 1.0:
        return math.pi / 2
    return math.asin(value)


def invSqrt(value: float) -> float:
    """Find the fast inverse square root.

    Use value**-0.5 instead
    """
    # https://pizer.wordpress.com/2008/10/12/fast-inverse-square-root/
    three_halfs = 1.5
    x2 = value * 0.5
    y = value

    packed_y = struct.pack("f", y)
    i = struct.unpack("i", packed_y)[0]  # treat float's bytes as int
    i = 0x5F3759DF - (i >> 1)  # arithmetic with magic number
    packed_i = struct.pack("i", i)
    y = struct.unpack("f", packed_i)[0]  # treat int's bytes as float

    y = y * (three_halfs - (x2 * y * y))  # Newton's method
    return y


class RunningAverage:
    """Running Average with Variance.

    With help from chat.openai.com
    """

    def __init__(self, window_size):
        self.window_size = window_size
        if self.window_size <= 0:
            raise ValueError("Window size must be greater than zero.")
        self.window = deque(maxlen=window_size)
        self.sum = Vector3D(x=0.0, y=0.0, z=0.0)
        self.squared_sum = Vector3D(x=0.0, y=0.0, z=0.0)

    def update(self, value):
        """Update the running average with a new value."""
        self.len = len(self.window)
        if self.len < self.window_size:
            self.window.append(value)
            self.sum = self.sum + value
            # self.avg = self.sum / len(self.window)
            self.squared_sum = self.squared_sum + (value * value)
        else:
            old_value = self.window.popleft()
            self.window.append(value)
            self.sum = self.sum + value - old_value
            self.squared_sum = (
                self.squared_sum + (value * value) - (old_value * old_value)
            )

    @property
    def avg(self) -> float:
        """Average of the sequence."""
        return self.sum / self.len

    @property
    def var(self) -> float:
        """Variance of the sequence."""
        return (self.squared_sum - (self.sum * self.sum) / self.len) / self.len


def vector_angle2q(vec: Vector3D, angle: float = 0.0) -> Quaternion:
    """Create quaternion based on rotation around a vector."""
    _vec = copy(vec)
    _vec.normalize()
    halfAngle = angle * 0.5
    sinHalfAngle = math.sin(halfAngle)
    return Quaternion(
        w=math.cos(halfAngle),
        x=vec.x * sinHalfAngle,
        y=vec.y * sinHalfAngle,
        z=vec.z * sinHalfAngle,
    )


def q2rpy(q: Quaternion) -> Vector3D:
    """Quaternion to roll pitch yaw.

    chat.openai.com
    """
    wx = q.w * q.x
    yz = q.y * q.z
    xx = q.x * q.x
    yy = q.y * q.y
    zz = q.z * q.z
    wy = q.w * q.y
    xz = q.x * q.z
    wz = q.w * q.z
    xy = q.x * q.y

    # roll (x-axis rotation)
    sinr_cosp = 2.0 * (wx + yz)
    cosr_cosp = 1.0 - 2.0 * (xx + yy)
    roll = math.atan2(sinr_cosp, cosr_cosp)

    # pitch (y-axis rotation)
    sinp = 2.0 * (wy - xz)
    if abs(sinp) >= 1.0:
        pitch = math.copysign(math.pi / 2.0, sinp)  # use 90 degrees if out of range
    else:
        pitch = math.asin(sinp)

    # yaw (z-axis rotation)
    siny_cosp = 2.0 * (wz + xy)
    cosy_cosp = 1.0 - 2.0 * (yy + zz)
    yaw = math.atan2(siny_cosp, cosy_cosp)

    return Vector3D(x=roll, y=pitch, z=yaw)


def rpy2q(r, p: float = 0.0, y: float = 0.0) -> Quaternion:
    """Assume vector contains roll, pitch, yaw and convert to quaternion.

    chat.openai.com
    accel2q is q.y and q.z have wrong sign
    """
    if isinstance(r, Vector3D):
        roll = r.x
        pitch = r.y
        yaw = r.z
    elif isinstance(r, np.ndarray):
        if len(r) == 3:
            roll, pitch, yaw = r
    elif isinstance(r, numbers.Number):
        roll = r
        pitch = p
        yaw = y
    else:
        raise TypeError(f"Unsupported operand type for rpy2q: {type(r)}")

    cy2 = math.cos(yaw * 0.5)
    sy2 = math.sin(yaw * 0.5)
    cp2 = math.cos(pitch * 0.5)
    sp2 = math.sin(pitch * 0.5)
    cr2 = math.cos(roll * 0.5)
    sr2 = math.sin(roll * 0.5)

    w = cy2 * cp2 * cr2 + sy2 * sp2 * sr2
    x = cy2 * cp2 * sr2 - sy2 * sp2 * cr2
    y = sy2 * cp2 * sr2 + cy2 * sp2 * cr2
    z = sy2 * cp2 * cr2 - cy2 * sp2 * sr2

    return Quaternion(w, x, y, z)


def accel2rpy(acc) -> Vector3D:
    """Gravity to roll pitch yaw hen X forward, Y right, Z down."""
    if isinstance(acc, Vector3D):
        _acc = copy(acc)
    elif isinstance(acc, np.ndarray):
        if len(acc) == 3:
            _acc = Vector3D(acc)
    else:
        raise TypeError(f"Unsupported operand type for accel2rpy: {type(acc)}")

    _acc.normalize()

    roll = math.atan2(_acc.y, _acc.z)
    pitch = math.atan2(-_acc.x, math.sqrt(_acc.y * _acc.y + _acc.z * _acc.z))
    yaw = 0.0
    return Vector3D(roll, pitch, yaw)


def accel2q(acc) -> Quaternion:
    """Convert Accelerometer to Quaternion assuming no motion."""
    if isinstance(acc, Vector3D):
        _acc = copy(acc)
    elif isinstance(acc, np.ndarray):
        _acc = Vector3D(acc)
    else:
        raise TypeError(f"Unsupported operand type for accel2rpy: {type(acc)}")

    _acc.normalize()
    # Calculate roll and pitch angles
    roll = math.atan2(_acc.y, _acc.z)
    # verify the negative sign on the pitch
    pitch = -math.atan2(-_acc.x, math.sqrt(_acc.y * _acc.y + _acc.z * _acc.z))
    # yaw   =  0.0

    cp2 = math.cos(pitch * 0.5)
    sp2 = math.sin(pitch * 0.5)
    cr2 = math.cos(roll * 0.5)
    sr2 = math.sin(roll * 0.5)
    # cy2 = 1.
    # sy2 = 0.

    w = cr2 * cp2
    x = sr2 * cp2
    y = -cr2 * sp2
    z = sr2 * sp2

    q = Quaternion(w, x, y, z)
    q.normalize()

    return q


def accelmag2rpy(acc, mag) -> Vector3D:
    """Estimate Pose Vector from Accelerometer and Compass.

    Assuming X forward, Y right, Z down
    1) Acceleration to Roll Pitch Yaw=0.0
    2) Update estimated Yaw in Euler and return result

    """
    if isinstance(acc, Vector3D):
        _acc = copy(acc)
    elif isinstance(acc, np.ndarray):
        if len(acc) == 3:
            _acc = Vector3D(acc)
    else:
        raise TypeError(f"Unsupported operand type for accel2rpy: {type(acc)}")

    if isinstance(mag, Vector3D):
        _mag = copy(mag)
    elif isinstance(mag, np.ndarray):
        if len(mag) == 3:
            _mag = Vector3D(mag)
    else:
        raise TypeError(f"Unsupported operand type for accel2rpy: {type(mag)}")

    # 1) calculate roll and pitch from acceleration (gravity) vector
    _acc.normalize()
    roll = math.atan2(_acc.y, _acc.z)
    pitch = math.atan2(-_acc.x, math.sqrt(_acc.y * _acc.y + _acc.z * _acc.z))

    # 2) calculate yaw from magnetometer (rotation around z-axis)
    _mag.normalize()
    mag_x = (
        _mag.x * math.cos(pitch)
        + _mag.y * math.sin(pitch) * math.sin(roll)
        + _mag.z * math.sin(pitch) * math.cos(roll)
    )
    mag_y = _mag.y * math.cos(roll) - _mag.z * math.sin(roll)
    yaw = math.atan2(-mag_y, mag_x)

    return Vector3D(roll, pitch, yaw)


def accelmag2q(acc, mag) -> Quaternion:
    """Estimate Pose Vector from Accelerometer and Compass.

    Assuming X forward, Y right, Z down

    Using the following approach:
    1) accel mag to rpy
    2) rpy to quaternion

    Compared to other approaches:

    - pypi AHRS
      R = am2DCM(a, m, frame=NED)
      q = dcm2quat(R)

        if frame.upper() not in ['ENU', 'NED']:
            raise ValueError("Wrong coordinate frame. Try 'ENU' or 'NED'")
            a = np.array(a)
            m = np.array(m)
            H = np.cross(m, a)
            H /= np.linalg.norm(H)
            a /= np.linalg.norm(a)
            M = np.cross(a, H)
            if frame.upper() == 'ENU':
                return np.array([[H[0], M[0], a[0]],
                                [H[1], M[1], a[1]],
                                [H[2], M[2], a[2]]])

            return np.array([[M[0], H[0], -a[0]],
                            [M[1], H[1], -a[1]],
                            [M[2], H[2], -a[2]]])

        if R.shape[0] != R.shape[1]:
            raise ValueError('Input is not a square matrix')
        if R.shape[0] != 3:
            raise ValueError('Input needs to be a 3x3 array or matrix')
        q = np.array([1., 0., 0., 0.])
        q[0] = 0.5*np.sqrt(1.0 + R.trace())
        q[1] = (R[1, 2] - R[2, 1]) / q[0]
        q[2] = (R[2, 0] - R[0, 2]) / q[0]
        q[3] = (R[0, 1] - R[1, 0]) / q[0]
        q[1:] /= 4.0
        return q / np.linalg.norm(q)

    - chat.openai.com
        Could not make work
    """
    if isinstance(acc, Vector3D):
        _acc = copy(acc)
    elif isinstance(acc, np.ndarray):
        if len(acc) == 3:
            _acc = Vector3D(acc)
    else:
        raise TypeError(f"Unsupported operand type for accel2rpy: {type(acc)}")

    if isinstance(mag, Vector3D):
        _mag = copy(mag)
    elif isinstance(mag, np.ndarray):
        if len(mag) == 3:
            _mag = Vector3D(mag)
    else:
        raise TypeError(f"Unsupported operand type for accel2rpy: {type(mag)}")

    # 1) calculate roll and pitch from acceleration (gravity) vector
    _acc.normalize()
    roll = math.atan2(_acc.y, _acc.z)
    pitch = math.atan2(-_acc.x, math.sqrt(_acc.y * _acc.y + _acc.z * _acc.z))

    # 2) calculate yaw from magnetometer (rotation around z-axis)
    _mag.normalize()
    mag_x = (
        _mag.x * math.cos(pitch)
        + _mag.y * math.sin(pitch) * math.sin(roll)
        + _mag.z * math.sin(pitch) * math.cos(roll)
    )
    mag_y = _mag.y * math.cos(roll) - _mag.z * math.sin(roll)
    yaw = math.atan2(-mag_y, mag_x)

    # 3) convert roll pitch yaw to quaternion
    cy2 = math.cos(yaw * 0.5)
    sy2 = math.sin(yaw * 0.5)
    cp2 = math.cos(pitch * 0.5)
    sp2 = math.sin(pitch * 0.5)
    cr2 = math.cos(roll * 0.5)
    sr2 = math.sin(roll * 0.5)

    w = cy2 * cp2 * cr2 + sy2 * sp2 * sr2
    x = cy2 * cp2 * sr2 - sy2 * sp2 * cr2
    y = sy2 * cp2 * sr2 + cy2 * sp2 * cr2
    z = sy2 * cp2 * cr2 - cy2 * sp2 * sr2

    q = Quaternion(w=w, x=x, y=y, z=z)

    return q


def rpymag2h(rpy: Vector3D, mag, declination=0.0) -> float:
    """Tilt compensated heading from compass corrected for local magnetic declination.

    Input:
      pose:       rpy
      mag:        Vector3D
      declination float
    Output:
      heading:    float
    """
    if isinstance(mag, Vector3D):
        _mag = copy(mag)
    elif isinstance(mag, np.ndarray):
        _mag = Vector3D(mag)
    else:
        raise TypeError(f"Unsupported operand type for mag: {type(mag)}")

    _mag.normalize()

    roll = rpy.x
    pitch = rpy.y

    tilted_mag_x = _mag.x * math.cos(pitch) + _mag.z * math.sin(pitch)
    tilted_mag_y = (
        _mag.x * math.sin(roll) * math.sin(pitch)
        + _mag.y * math.cos(roll)
        - _mag.z * math.sin(roll) * math.cos(pitch)
    )

    heading = math.atan2(-tilted_mag_y, tilted_mag_x) + declination

    return heading if heading > 0.0 else 2 * np.pi + heading


def qmag2h(q: Quaternion, mag, declination=0.0) -> float:
    """Tilt compensated heading from compass corrected for local magnetic declination.

    Input:
      pose:       Quaternion
      mag:        Vector3D
      declination float
    Output:
      heading:    float
    """
    if isinstance(mag, Vector3D):
        _mag = copy(mag)
    elif isinstance(mag, np.ndarray):
        _mag = Vector3D(mag)
    else:
        raise TypeError(f"Unsupported operand type for mag: {type(mag)}")

    wx = q.w * q.x
    yz = q.y * q.z
    xx = q.x * q.x
    yy = q.y * q.y
    wy = q.w * q.y
    xz = q.x * q.z

    # roll
    sinr_cosp = 2.0 * (wx + yz)
    cosr_cosp = 1.0 - 2.0 * (xx + yy)
    roll = math.atan2(sinr_cosp, cosr_cosp)

    # pitch
    sinp = 2.0 * (wy - xz)
    if abs(sinp) >= 1.0:
        pitch = math.copysign(math.pi / 2.0, sinp)  # use 90 degrees if out of range
    else:
        pitch = math.asin(sinp)

    _mag.normalize()

    tilted_mag_x = _mag.x * math.cos(pitch) + _mag.z * math.sin(pitch)
    tilted_mag_y = (
        _mag.x * math.sin(roll) * math.sin(pitch)
        + _mag.y * math.cos(roll)
        - _mag.z * math.sin(roll) * math.cos(pitch)
    )

    heading = math.atan2(-tilted_mag_y, tilted_mag_x) + declination

    return heading if heading > 0.0 else 2 * np.pi + heading


def q2gravity(pose: Quaternion) -> Vector3D:
    """Create unit Gravity vector from pose quaternion.

    North East Down (Gravity is positive in Z direction pointing down)

    3x3 rotation matrix from quaternion
    multiply with dot product

    ([
        [1.0 - 2.*(yy + zz),          2.*(xy - zw),          2.*(xz + yw)],
        [      2.*(xy + zw),    1.0 - 2.*(xx + zz),          2.*(yz - xw)],
        [      2.*(xz - yw),          2.*(yz + xw),    1.0 - 2.*(xx + yy)]
    ])

    We rotate from world to frame; therefore, the rotation matrix is transposed

    Rotated_vector = np.dot(r33.T, np.array([0,0,1]))

    """
    x = 2.0 * (pose.x * pose.z - pose.w * pose.y)
    y = 2.0 * (pose.y * pose.z + pose.w * pose.x)
    z = 1.0 - 2.0 * (pose.x * pose.x + pose.y * pose.y)

    return Vector3D(x, y, z)


def sensorAcc(acc: Vector3D, q: Quaternion, g: float) -> Vector3D:
    """Compute residual acceleration in sensor frame."""
    return acc - g * q2gravity(q)


def earthAcc(acc: Vector3D, q: Quaternion, g: float) -> Vector3D:
    """Compute residual acceleration in earth frame."""
    # acc_r = (q * acc * q.conjugate).v
    acc_r = acc.rotate(q.r33)
    acc_r.z = acc_r.z - g  # subtract gravity
    return acc_r


def gravity(latitude: float, altitude: float) -> float:
    """Gravity on Ellipsoid Surface.

    from https://github.com/Mayitzin/ahrs/blob/master/ahrs/utils/wgs84.py
    """
    a = 6_378_137.0  # EARTH_EQUATOR_RADIUS
    f = 1.0 / 298.257223563  # EARTH_FLATTENING
    gm = 3.986004418e14  # EARTH_GM
    w = 7.292115e-5  # EARTH_ROTATION
    b = a * (1 - f)  # EARTH_POLE_RADIUS
    ge = 9.78032533590406  # EARTH_EQUATOR_GRAVITY
    gp = 9.832184937863065  # EARTH_POLE_GRAVITY
    e2 = 0.0066943799901413165  # EARTH_ECCENTRICITY_SQUARED
    lat = np.deg2rad(latitude)  # LATITUDE in radians
    k = (b * gp) / (a * ge) - 1
    sin2 = math.sin(lat) ** 2
    gravity = (
        ge * (1 + k * sin2) / math.sqrt(1 - e2 * sin2)
    )  # Gravity on Ellipsoid Surface
    if altitude != 0.0:
        m = w**2 * a**2 * b / gm  # Gravity constant
        gravity *= (
            1.0
            - 2.0 * altitude * (1.0 + f + m - 2 * f * sin2) / a
            + 3.0 * altitude**2 / a**2
        )  # Gravity Above Ellipsoid

    return gravity



"""Quaternion and Vector3D data types and functions."""

import math
import numbers

import numpy as np
from numpy.typing import NDArray

from py_imu.definitions import EPSILON


class Quaternion:
    """Quaternion Class.

    q1 = Quaternion(1., 2., 3., 4.)
    q2 = Quaternion(w=5., x=6., y=7., z=8.)
    q3 = Quaternion(np.array([9,10,11,12]))

    q5 = q1 + q2
    q6 = q1 * q2
    q7 = 2 * q1

    q1.conjugate
    q1.inverse
    q1.normalize()
    q1.r33 (cosine matrix)
    q1.norm: length of quaternion
    q1.v: vector part of quaternion as Vector3D
    q1.q: quaternion as np.array
    """

    def __init__(self, w=0.0, x=0.0, y=0.0, z=0.0, v=None):
        # allows any possible combination to be passed in
        if v is None:
            if isinstance(w, numbers.Number):
                self.w = w
                self.x = x
                self.y = y
                self.z = z

    def __str__(self):
        return f"Quaternion({self.w}, {self.x}, {self.y}, {self.z})"

    def __repr__(self):
        return str(self)

    def __add__(self, other):
        """Add two quaternions or quaternion and scalar."""
        if isinstance(other, Quaternion):
            return Quaternion(
                self.w + other.w, self.x + other.x, self.y + other.y, self.z + other.z
            )
        elif isinstance(other, np.ndarray):
            return Quaternion(
                self.w + other[0],
                self.x + other[1],
                self.y + other[2],
                self.z + other[3],
            )
        elif isinstance(other, numbers.Number):
            return Quaternion(
                self.w + other, self.x + other, self.y + other, self.z + other
            )
        else:
            raise TypeError(
                f"Unsupported operand type for +: Quaternion and {type(other)}"
            )

    def __sub__(self, other):
        """Subtract two quaternions or quaternion and scalar."""
        if isinstance(other, Quaternion):
            return Quaternion(
                self.w - other.w, self.x - other.x, self.y - other.y, self.z - other.z
            )
        elif isinstance(other, np.ndarray):
            return Quaternion(
                self.w - other[0],
                self.x - other[1],
                self.y - other[2],
                self.z - other[3],
            )
        elif isinstance(other, numbers.Number):
            return Quaternion(
                self.w - other, self.x - other, self.y - other, self.z - other
            )
        else:
            raise TypeError(
                f"Unsupported operand type for -: Quaternion and {type(other)}"
            )

    def __mul__(self, other):
        """Multiply two quaternions or quaternion and vector."""
        if isinstance(other, Quaternion):
            w = (
                (self.w * other.w)
                - (self.x * other.x)
                - (self.y * other.y)
                - (self.z * other.z)
            )
            x = (
                (self.w * other.x)
                + (self.x * other.w)
                + (self.y * other.z)
                - (self.z * other.y)
            )
            y = (
                (self.w * other.y)
                - (self.x * other.z)
                + (self.y * other.w)
                + (self.z * other.x)
            )
            z = (
                (self.w * other.z)
                + (self.x * other.y)
                - (self.y * other.x)
                + (self.z * other.w)
            )
            return Quaternion(w, x, y, z)
        elif isinstance(other, Vector3D):
            """
            multiply quaternion with vector
            vector is converted to quaternion with [0,vector]
            then computed the same as above with other.w=0
            """
            w = -(self.x * other.x) - (self.y * other.y) - (self.z * other.z)
            x = self.w * other.x + self.y * other.z - self.z * other.y
            y = self.w * other.y - self.x * other.z + self.z * other.x
            z = self.w * other.z + self.x * other.y - self.y * other.x
            return Quaternion(w, x, y, z)
        elif isinstance(other, numbers.Number):
            """multiply with scalar"""
            return Quaternion(
                self.w * other, self.x * other, self.y * other, self.z * other
            )
        else:
            raise TypeError("Unsupported operand type")

    def __rmul__(self, other):
        """Right multiplication."""
        return self.__mul__(other)

    def __truediv__(self, other):
        """Division."""
        if isinstance(other, numbers.Number):
            return Quaternion(
                self.w / other, self.x / other, self.y / other, self.z / other
            )
        else:
            raise TypeError("Unsupported operand type")

    def __floordiv__(self, other):
        """Floor division."""
        if isinstance(other, numbers.Number):
            return Quaternion(
                self.w // other, self.x // other, self.y // other, self.z // other
            )
        else:
            raise TypeError("Unsupported operand type")

    def __eq__(self, other):
        """Check that the two quaternions are equal."""
        return (
            self.w == other.w
            and self.x == other.x
            and self.y == other.y
            and self.z == other.z
        )

    def normalize(self):
        """Normalize the quaternion."""
        mag = self.norm
        if mag != 0:
            self.w = self.w / mag
            self.x = self.x / mag
            self.y = self.y / mag
            self.z = self.z / mag

    @property
    def v(self) -> NDArray:
        """Extract the vector component of the quaternion."""
        # return np.array([self.x,self.y,self.z])
        return Vector3D(self.x, self.y, self.z)

    @property
    def q(self) -> NDArray:
        """Convert the quaternion to np.array."""
        # return np.array([self.x,self.y,self.z])
        return np.array([self.w, self.x, self.y, self.z])

    @property
    def conjugate(self):
        """Conjugate of quaternion."""
        return Quaternion(self.w, -self.x, -self.y, -self.z)

    @property
    def norm(self) -> float:
        """Length of quaternion."""
        return math.sqrt(
            self.w * self.w + self.x * self.x + self.y * self.y + self.z * self.z
        )

    @property
    def inverse(self):
        """Inverse of quaternion."""
        return self.conjugate / self.norm

    @property
    def r33(self) -> NDArray:
        """Quaternion to 3x3 rotation matrix."""
        # Normalize quaternion
        # self.normalize()

        # Compute rotation matrix elements
        xx = self.x * self.x
        xy = self.x * self.y
        xz = self.x * self.z
        xw = self.x * self.w
        yy = self.y * self.y
        yz = self.y * self.z
        yw = self.y * self.w
        zz = self.z * self.z
        zw = self.z * self.w

        # Construct the rotation matrix
        return np.array(
            [
                [1.0 - 2.0 * (yy + zz), 2.0 * (xy - zw), 2.0 * (xz + yw)],
                [2.0 * (xy + zw), 1.0 - 2.0 * (xx + zz), 2.0 * (yz - xw)],
                [2.0 * (xz - yw), 2.0 * (yz + xw), 1.0 - 2.0 * (xx + yy)],
            ]
        )

    @property
    def isZero(self) -> bool:
        """Check whether the quaternion is zero."""
        return (
            abs(self.w) <= EPSILON
            and abs(self.x) <= EPSILON
            and abs(self.y) <= EPSILON
            and abs(self.z) <= EPSILON
        )


class Vector3D:
    """3D Vector Class.

    v1 = Vector3D(1., 2., 3.)
    v2 = Vector3D(x=4., y=5., z=6.)
    v3 = Vector3D(np.array([7,8,9]))

    v4 = v1 + v2
    v5 = v1 * v2
    v6 = 2. * v1

    v1.dot(v2)
    v1.cross(v2)
    v1.normalize()
    v1.norm: length of vector
    v1.rotate(np.array[3x3])
    v1.v: vector as np.array
    v1.q: vector as quaternion with w=0.
    """

    def __init__(self, x=0.0, y=0.0, z=0.0):
        if isinstance(x, numbers.Number):
            self.x = x
            self.y = y
            self.z = z
        elif isinstance(x, np.ndarray):
            if len(x) == 3:
                self.x = x[0]
                self.y = x[1]
                self.z = x[2]
        elif isinstance(x, Vector3D):
            self.x = x.x
            self.y = x.y
            self.z = x.z

    def __copy__(self):
        """Copy the vector."""
        return Vector3D(self.x, self.y, self.z)

    def __bool__(self):
        """Check whether the vector is non-zero."""
        return not self.isZero

    def __abs__(self):
        """Absolute value of the vector."""
        return Vector3D(abs(self.x), abs(self.y), abs(self.z))

    def __neg__(self):
        """Negate the vector."""
        return Vector3D(-self.x, -self.y, -self.z)

    def __len__(self):
        """Length of the vector."""
        return 3

    def __str__(self):
        """Representation of the vector."""
        return f"Vector3D({self.x}, {self.y}, {self.z})"

    def __repr__(self):
        """Representation of the vector."""
        return str(self)

    def __add__(self, other):
        """Add two vectors."""
        if isinstance(other, Vector3D):
            return Vector3D(self.x + other.x, self.y + other.y, self.z + other.z)
        elif isinstance(other, numbers.Number):
            return Vector3D(self.x + other, self.y + other, self.z + other)
        else:
            raise TypeError(
                f"Unsupported operand type for +: Vector3D and {type(other)}"
            )

    def __sub__(self, other):
        """Subtract two vectors."""
        if isinstance(other, Vector3D):
            return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)
        elif isinstance(other, numbers.Number):
            return Vector3D(self.x - other, self.y - other, self.z - other)
        else:
            raise TypeError(
                f"Unsupported operand type for -: Vector3D and {type(other)}"
            )

    def __mul__(self, other):
        """Multiplication."""
        if isinstance(other, numbers.Number):
            return Vector3D(self.x * other, self.y * other, self.z * other)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x * other.x, self.y * other.y, self.z * other.z)
        elif isinstance(other, np.ndarray):
            shape = other.shape
            if len(shape) == 1:
                if shape[0] == 3:
                    return Vector3D(
                        self.x * other[0], self.y * other[1], self.z * other[2]
                    )
                elif shape[0] == 4:
                    # other is quaternion of w,x,y,z convert vector to quaternion [0,x,y,z]

                    # x1, y1, z1     = self.v # w1 is 0, dont use
                    other_w, other_x, other_y, other_z = other
                    w = -self.x * other_x - self.y * other_y - self.z * other_z
                    x = self.x * other_w + self.y * other_z - self.z * other_y
                    y = -self.x * other_z + self.y * other_w + self.z * other_x
                    z = self.x * other_y - self.y * other_x + self.z * other_w
                    return Quaternion(w, x, y, z)

            elif shape == (3, 3):
                """Matrix Multiplication"""
                rotated_vector = np.dot(other, np.array([self.x, self.y, self.z]))
                # rotated_vector = np.dot(np.array([self.x,self.y,self.z]),other)
                return Vector3D(
                    x=rotated_vector[0], y=rotated_vector[1], z=rotated_vector[2]
                )

        elif isinstance(other, Quaternion):
            w = -self.x * other.x - self.y * other.y - self.z * other.z
            x = self.x * other.w + self.y * other.z - self.z * other.y
            y = -self.x * other.z + self.y * other.w + self.z * other.x
            z = self.x * other.y - self.y * other.x + self.z * other.w
            return Quaternion(w, x, y, z)
        else:
            raise TypeError(
                f"Unsupported operand type for *: Vector3D and {type(other)}"
            )

    def __rmul__(self, other):
        """Right multiplication."""
        return self.__mul__(other)

    def __truediv__(self, other):
        """Division."""
        if isinstance(other, numbers.Number):
            return Vector3D(self.x / other, self.y / other, self.z / other)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x / other.x, self.y / other.y, self.z / other.z)
        else:
            raise ValueError(f"Unsupported operand type for /: {other}")

    def __floordiv__(self, other):
        """Floor division."""
        if isinstance(other, numbers.Number):
            return Vector3D(self.x // other, self.y // other, self.z // other)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x // other.x, self.y // other.y, self.z // other.z)
        else:
            raise ValueError(f"Unsupported operand type for //: {other}")

    def __pow__(self, other):
        """Potentiate."""
        if isinstance(other, numbers.Number):
            return Vector3D(self.x**other, self.y**other, self.z**other)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x**other.x, self.y**other.y, self.z**other.z)
        else:
            raise TypeError(
                f"Unsupported operand type for **: Vector3D and {type(other)}"
            )

    def __eq__(self, other):
        """Check whether the vectors are equal."""
        if isinstance(other, numbers.Number):
            return Vector3D(self.x == other, self.y == other, self.z == other)
        elif isinstance(other, Vector3D):
            return Vector3D(self.x == other.x, self.y == other.y, self.z == other.z)
        else:
            raise TypeError(
                f"Unsupported operand type for ==: Vector3D and {type(other)}"
            )

    def __lt__(self, other):
        """Is vector smaller than the other."""
        if isinstance(other, numbers.Number):
            return Vector3D(self.x < other, self.y < other, self.z < other)
        elif isinstance(other, Vector3D):
            return Vector3D(x=self.x < other.x, y=self.y < other.y, z=self.z < other.z)
        else:
            raise TypeError(
                f"Unsupported operand type for <=: Vector3D and {type(other)}"
            )

    def normalize(self):
        """Normalize vector."""
        mag = math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)
        if mag != 0:
            self.x = self.x / mag
            self.y = self.y / mag
            self.z = self.z / mag

    def rotate(self, other):
        """Rotate vector by rotation matrix."""
        if isinstance(other, np.ndarray):
            if other.shape == (3, 3):
                rotated_vector = np.dot(other, np.array([self.x, self.y, self.z]))
                # rotated_vector = np.dot(np.array([self.x,self.y,self.z]),other)
                return Vector3D(
                    x=rotated_vector[0], y=rotated_vector[1], z=rotated_vector[2]
                )
            else:
                raise TypeError(
                    f"Unsupported operand type for cross product: Vector3D and nd.array of shape {other.shape}"
                )
        else:
            raise TypeError(
                f"Unsupported operand type for cross product: Vector3D and {type(other)}"
            )

    @property
    def q(self):
        """Return np array with rotation 0 and vector x, y, z."""
        return Quaternion(w=0, x=self.x, y=self.y, z=self.z)

    @property
    def v(self) -> NDArray:
        """Returns np array of vector."""
        return np.array([self.x, self.y, self.z])

    @property
    def isZero(self) -> bool:
        """Check if vector is zero."""
        return (
            abs(self.x) <= EPSILON and abs(self.y) <= EPSILON and abs(self.z) <= EPSILON
        )



"""Configure the logger."""

import sys
from datetime import datetime
from pathlib import Path

from loguru import logger

from py_imu.definitions import (
    DATE_FORMAT,
    DEFAULT_LOG_FILENAME,
    DEFAULT_LOG_LEVEL,
    ENCODING,
    LOG_DIR,
)


def create_timestamped_filepath(suffix: str, output_dir: Path, prefix: str) -> Path:
    """Generate a timestamped filename.

    :param suffix: Suffix to append to the timestamped filename.
    :param output_dir: Output directory.
    :param prefix: Prefix to append to the timestamped filename.
    :return: Path to the timestamped filename.
    """
    timestamp = datetime.now().strftime(DATE_FORMAT)
    filepath = output_dir / f"{prefix}_{timestamp}.{suffix}"
    filepath.parent.mkdir(parents=True, exist_ok=True)  # create dirs if missing
    filepath.touch(exist_ok=True)  # create empty file (don't overwrite)
    return filepath


def setup_logger(
    filename: str = DEFAULT_LOG_FILENAME,
    stderr_level: str = DEFAULT_LOG_LEVEL,
    log_level: str = DEFAULT_LOG_LEVEL,
    log_dir: Path | None = None,
) -> Path:
    """Configure the logger.

    :param filename: Name of the file to create.
    :param stderr_level: Logging level to use.
    :param log_level: Logging level to use.
    :param log_dir: Logging directory to use.
    :return: Path to the created logfile.
    """
    logger.remove()

    if log_dir is None:
        log_filepath = LOG_DIR
    else:
        log_filepath = log_dir
    filepath_with_time = create_timestamped_filepath(
        output_dir=log_filepath, prefix=filename, suffix="log"
    )
    logger.add(sys.stderr, level=stderr_level)
    logger.add(filepath_with_time, level=log_level, encoding=ENCODING, enqueue=True)
    logger.info(f"Logging to '{filepath_with_time}'.")
    return filepath_with_time



"""Common definitions."""

import math
from dataclasses import asdict, dataclass
from pathlib import Path

import numpy as np

np.set_printoptions(precision=3, floatmode="fixed", suppress=True)


TWOPI = 2.0 * math.pi
PI_HALF = math.pi / 2.0
EPSILON = 2.0 * math.ldexp(1.0, -53)


# --- Directories ---
ROOT_DIR: Path = Path("src").parent
DATA_DIR: Path = ROOT_DIR / "data"
RECORDINGS_DIR: Path = DATA_DIR / "recordings"
LOG_DIR: Path = DATA_DIR / "logs"

# Default encoding
ENCODING: str = "utf-8"

DATE_FORMAT = "%Y-%m-%d_%H-%M-%S"

DUMMY_VARIABLE = "dummy_variable"


@dataclass
class LogLevel:
    """Log level."""

    trace: str = "TRACE"
    debug: str = "DEBUG"
    info: str = "INFO"
    success: str = "SUCCESS"
    warning: str = "WARNING"
    error: str = "ERROR"
    critical: str = "CRITICAL"

    def __iter__(self):
        """Iterate over log levels."""
        return iter(asdict(self).values())


DEFAULT_LOG_LEVEL = LogLevel.info
DEFAULT_LOG_FILENAME = "log_file"



"""Pose and Motion Estimation with Madgwick Filter."""

import math
from copy import copy
from typing import Optional

import numpy as np

from py_imu.quaternion import Quaternion, Vector3D
from py_imu.utilities import accel2q, accelmag2q

TWOPI = 2.0 * math.pi
IDENTITY_QUATERNION = Quaternion(1.0, 0.0, 0.0, 0.0)
VECTOR_ZERO = Vector3D(0.0, 0.0, 0.0)
DEG2RAD = math.pi / 180.0
RAD2DEG = 180.0 / math.pi
EPSILON = math.ldexp(1.0, -53)


def updateIMU(
    q: Quaternion, gyr: Vector3D, acc: Vector3D, dt: float, gain: float
) -> Quaternion:
    """Quaternion Estimation with a Gyroscope and Accelerometer.

    q   : A-priori quaternion.
    gyr : Vector3D of tri-axial Gyroscope in rad/s
    acc : Vector3D of tri-axial Accelerometer in m/s^2
    dt  : float, default: None, Time step, in seconds, between consecutive Quaternions.

    Returns
    -------
    q   : Estimated quaternion.

    """
    acc.normalize()  #          // 160-162
    # q.normalize() its normalized at the end

    # Estimated orientation change from gyroscope
    qDot = 0.5 * (q * gyr)  # (eq. 12) // 150-153

    # Objective function                                       # (eq. 25)
    f = np.array(
        [
            2.0 * (q.x * q.z - q.w * q.y) - acc.x,
            2.0 * (q.w * q.x + q.y * q.z) - acc.y,
            2.0 * (0.5 - q.x * q.x - q.y * q.y) - acc.z,
        ]
    )

    # if np.linalg.norm(f) > 0:
    if math.sqrt(f[0] * f[0] + f[1] * f[1] + f[2] * f[2]) > 0:
        # Jacobian                                             # (eq. 26)
        J = np.array(
            [
                [-2.0 * q.y, 2.0 * q.z, -2.0 * q.w, 2.0 * q.x],
                [2.0 * q.x, 2.0 * q.w, 2.0 * q.z, 2.0 * q.y],
                [0.0, -4.0 * q.x, -4.0 * q.y, 0.0],
            ]
        )

        # Sensitivity Matrix                                   # (eq. 34)
        gradient = J.T @ f

        # gradient = gradient / np.linalg.norm(gradient)         #           // 184-188
        gradient = gradient / math.sqrt(
            gradient[0] * gradient[0]
            + gradient[1] * gradient[1]
            + gradient[2] * gradient[2]
        )

        # Update orientation change
        qDot = qDot - gain * gradient  # (eq. 33) // 191-194

    # Update orientation
    q = q + qDot * dt  # (eq. 13) // 198-201
    q.normalize()  #          // 204-208

    return q


def updateMARG(
    q: Quaternion,
    sensor_data: tuple[Vector3D, Vector3D, Vector3D],
    dt: float,
    gain: float,
) -> Quaternion:
    """Quaternion Estimation with a Gyroscope, Accelerometer and Magnetometer.

    q   : A-priori quaternion.
    gyr : Vector3D of tri-axial Gyroscope in rad/s
    acc : Vector3D of tri-axial Accelerometer in m/s^2
    mag : Vector3D of tri-axial Magnetometer in nT
    dt  : float, default: None, Time step, in seconds, between consecutive Quaternions.

    Returns
    -------
    q : Estimated quaternion.

    """
    gyr, acc, mag = sensor_data

    acc.normalize()
    mag.normalize()
    # q.normalize() its normalized at the end

    # Estimated orientation change from gyroscope
    qDot = 0.5 * (q * gyr)  # (eq. 12)

    # Rotate normalized magnetometer measurements
    # h = mag.rotate(q.r33.T)                                  # (eq. 45) , 19 microseconds
    h = q * mag * q.conjugate  # (eq. 45) , 14 microseconds
    bx = math.sqrt(h.x * h.x + h.y * h.y)  # (eq. 46)
    bz = h.z

    # Objective function                                       # (eq. 31)
    f = np.array(
        [
            2.0 * (q.x * q.z - q.w * q.y) - acc.x,
            2.0 * (q.w * q.x + q.y * q.z) - acc.y,
            2.0 * (0.5 - q.x * q.x - q.y * q.y) - acc.z,
            2.0 * bx * (0.5 - q.y * q.y - q.z * q.z)
            + 2.0 * bz * (q.x * q.z - q.w * q.y)
            - mag.x,
            2.0 * bx * (q.x * q.y - q.w * q.z)
            + 2.0 * bz * (q.w * q.x + q.y * q.z)
            - mag.y,
            2.0 * bx * (q.w * q.y + q.x * q.z)
            + 2.0 * bz * (0.5 - q.x * q.x - q.y * q.y)
            - mag.z,
        ]
    )

    # if np.linalg.norm(f) > 0:                                  # math.sqrt(f[0]*f[0] + f[1]*f[1] + ... ) is faster
    if (
        math.sqrt(
            f[0] * f[0]
            + f[1] * f[1]
            + f[2] * f[2]
            + f[3] * f[3]
            + f[4] * f[4]
            + f[5] * f[5]
        )
        > 0
    ):
        # Jacobian                                             # eq. 32)
        J = np.array(
            [
                [-2.0 * q.y, 2.0 * q.z, -2.0 * q.w, 2.0 * q.x],
                [2.0 * q.x, 2.0 * q.w, 2.0 * q.z, 2.0 * q.y],
                [0.0, -4.0 * q.x, -4.0 * q.y, 0.0],
                [
                    -2.0 * bz * q.y,
                    2.0 * bz * q.z,
                    -4.0 * bx * q.y - 2.0 * bz * q.w,
                    -4.0 * bx * q.z + 2.0 * bz * q.x,
                ],
                [
                    -2.0 * bx * q.z + 2.0 * bz * q.x,
                    2.0 * bx * q.y + 2.0 * bz * q.w,
                    2.0 * bx * q.x + 2.0 * bz * q.z,
                    -2.0 * bx * q.w + 2.0 * bz * q.y,
                ],
                [
                    2.0 * bx * q.y,
                    2.0 * bx * q.z - 4.0 * bz * q.x,
                    2.0 * bx * q.w - 4.0 * bz * q.y,
                    2.0 * bx * q.x,
                ],
            ]
        )

        # Sensitivity Matrix
        gradient = J.T @ f  # (eq. 34)
        # gradient = gradient / np.linalg.norm(gradient)
        gradient = gradient / math.sqrt(
            gradient[0] * gradient[0]
            + gradient[1] * gradient[1]
            + gradient[2] * gradient[2]
            + gradient[3] * gradient[3]
        )
        # Updated orientation change
        qDot = qDot - gain * gradient  # (eq. 33)

    # Update orientation
    q = q + qDot * dt  # (eq. 13)
    q.normalize()
    return q


class Madgwick:
    """Madgwick's Gradient Descent Pose Filter.

    Earth Axis Convention: NED (North, East, Down)

    Methods:
    - update:
      Update the filter with new measurements, calls
      updateIMU when using only gyroscope and accelerometer data (IMU implementation),
      updateMARG when using gyroscope, accelerometer and magnetometer data (MARG implementation).

    Initialization:
      frequency : float, default: 100.0; Sampling frequency in Hertz, or
      dt : float, default: 0.01;         Sampling step in seconds. Inverse of sampling frequency. Not required.
      gain : float,                      Filter gain. Defaults to 0.033 for IMU implementations, or to 0.041 for MARG implementations.
      gain_imu : float, default: 0.033;  Filter gain for IMU implementation.
      gain_marg : float, default: 0.041; Filter gain for MARG implementation.

    """

    def __init__(self, **kwargs):
        self.q = None
        self.acc = None
        self.gyr = None
        self.mag = None
        self.frequency: float = kwargs.get("frequency", 100.0)
        self.dt: float = kwargs.get(
            "dt", (1.0 / self.frequency) if self.frequency else 0.01
        )
        self.gain_imu = kwargs.get("gain_imu", 0.033)
        self.gain_marg = kwargs.get("gain_marg", 0.041)

    def update(
        self,
        gyr: Vector3D,
        acc: Vector3D,
        mag: Optional[Vector3D] = None,
        dt: float = -1,
    ) -> Quaternion:
        """Estimate the pose quaternion.

        gyr : Vector3D of Gyroscope in rad/s
        acc : Vector3D of Accelerometer in m/s^2
        mag : Vector3D of Magnetometer in nT, optional
        dt  : float, default: None, Time step, in seconds, between consecutive function calls.
        """
        self.gyr = copy(gyr)
        self.acc = copy(acc)

        if mag is None:
            # Compute with IMU architecture
            if (self.q is None) or (dt < 0):
                # We run this the first time. Estimate initial quaternion.
                #   Nake sure that you have stable readings from the senor before
                #   calling this function, otherwise it takes a while for the sensor to orient.
                self.q = accel2q(self.acc)  # estimate initial orientation
                # self.q.normalize()
                # print('Init with acc only: ', q2rpy(self.q))
            else:
                self.q = updateIMU(
                    self.q, self.gyr, self.acc, dt=dt, gain=self.gain_imu
                )

        else:
            # Compute with MARG architecture
            self.mag = copy(mag)
            if (self.q is None) or (dt < 0):
                # We run this the first time. Estimate initial quaternion.
                #   Nake sure that you have stable readings from the senor before
                #   calling this function, otherwise it will take a while for sensor to orient.
                self.q = accelmag2q(self.acc, self.mag)
                # self.q.normalize()
                # print('Init with acc and mag: ', q2rpy(self.q))
            else:
                sensor_data = (self.gyr, self.acc, self.mag)
                self.q = updateMARG(self.q, sensor_data, dt=dt, gain=self.gain_marg)

        return self.q
